#include <stdlib.h>
#include <string.h>
#include "bitops.h"
#include "huffctable.h"
#include "huffcoder.h"

// maximum hwidth - must be 1 <= X < 31
// higher mean more memory consumption and speed
#define MAX_HWIDTH 8
// bit buffer size - careful changing this!
// should not be smaller than 8 or MAX_HWIDTH
#define BIT_BUFFER_SIZE	16


/* -----------------------------------------------
	convert canonical to hcode structs
	----------------------------------------------- */
	
huffman_code** convert_ctable2hcode( canonical_table* ctable ) {
	huffman_code** table;
	huffman_code* tcode;
	unsigned char* value;
	unsigned char* nlen;
	unsigned int code;
	int ncodes;
	int len;
	int nl;
	
	
	// safety check
	if ( ctable == NULL ) return NULL;
	
	// build and preset huffman code table
	table = ( huffman_code** ) calloc( 256, sizeof( huffman_code* ) );
	if ( table == NULL ) return NULL;
	for ( int i = 0; i < 256; i++ ) table[i] = NULL;
	
	// setup tables
	if ( ctable->len_ptr == NULL ) { // incompressible case table setup
		// build codes - 1 to 1 symbol representation
		for ( code = 0; code < 256; code++ ) {
			tcode = ( huffman_code* ) calloc( 1, sizeof( huffman_code ) );
			if ( tcode == NULL ) { free( table ); return NULL; }
			tcode->len = 8;
			tcode->code = code;
			table[ code ] = tcode;
		}
	} else if ( ctable->n_symbols == 1 ) { // only one symbol in the table
		tcode = ( huffman_code* ) calloc( 1, sizeof( huffman_code ) );
		if ( tcode == NULL ) { free( table ); return NULL; }
		tcode->len = 0;
		tcode->code = 0;
		table[ *(ctable->val_ptr) ] = tcode;
	} else { // standard table setup
		// set up canonical table pointers
		nlen = ctable->len_ptr;
		value = ctable->val_ptr;
		ncodes = ctable->n_symbols;
		
		// build actual codes
		for ( len = 1, code = 0; ncodes > 0; len++ ) {
			for ( nl = (*nlen++); nl > 0; nl--, ncodes-- ) {
				tcode = ( huffman_code* ) calloc( 1, sizeof( huffman_code ) );
				if ( tcode == NULL ) { free( table ); return NULL; }
				tcode->len = len;
				tcode->code = code++;
				table[ (*value++) ] = tcode;
			}
			code <<= 1;
		}
	}
	
	
	return table;
}


/* -----------------------------------------------
	convert canonical to hconv recursive helper
	----------------------------------------------- */
	
huffman_conv_set* convert_ctable2hconv_rec( int* len_ptr, int** val_ptr, int mlen ) {
	huffman_conv_set* conv_set;
	huffman_conv** table;
	huffman_conv* tconv;
	unsigned int c0, c1;
	unsigned int r;
	int hwidth_l;
	int len;	
	int t;
	
	
	// find local hwidth and range
	if ( mlen <= MAX_HWIDTH ) hwidth_l = mlen;
	else for ( r = 1 << MAX_HWIDTH, t = MAX_HWIDTH, hwidth_l = 0;
		( r > 0 ) && ( hwidth_l < MAX_HWIDTH );
		r -= len_ptr[ hwidth_l++ ] * ( 1 << (--t) ) );
	r = 1 << hwidth_l;
	
	// found, allocate memory
	conv_set = ( huffman_conv_set* ) calloc( 1, sizeof( huffman_conv_set ) );
	table = ( huffman_conv** ) calloc( r, sizeof( huffman_conv* ) );
	if ( ( conv_set == NULL ) || ( table == NULL ) ) {
		if ( conv_set != NULL ) free( conv_set );
		if ( table != NULL ) free( table );
		return NULL;
	}
	
	// basic set up of conv_set struct
	for ( c0 = 0; c0 < r; c0++ ) table[ c0 ] = NULL;
	conv_set->hwidth = hwidth_l;
	conv_set->h = table;
	
	// actual generation of conversion table
	for ( len = 1, c0 = c1 = 0; c0 < r; ) {
		// find length
		for ( ; len_ptr[ len ] == 0; len++ );
		// generate new entry
		tconv = ( huffman_conv* ) calloc( 1, sizeof( huffman_conv ) );
		if ( tconv == NULL ) { free( conv_set ); return NULL; }
		
		if ( len <= hwidth_l ) { // routine for regular conversion table
			// set conversion attributes
			tconv->len = len;
			tconv->val = (**val_ptr);
			tconv->ext = NULL;
			// insert it into the table
			for ( c1 += 1 << ( hwidth_l - len ); c0 < c1; c0++ )
				table[ c0 ] = tconv;
			len_ptr[ len ]--;
			(*val_ptr)++;
		} else { // routine for ext conversion table
			// set conversion attributes
			tconv->len = hwidth_l;
			tconv->val = -1;
			tconv->ext = convert_ctable2hconv_rec( len_ptr + hwidth_l, val_ptr, mlen - hwidth_l );
			// insert it into the table
			for ( c1++; c0 < c1; c0++ )
				table[ c0 ] = tconv;
		}
	}
	
	
	return conv_set;
}
	
/* -----------------------------------------------
	convert canonical to hconv structs
	----------------------------------------------- */
	
huffman_conv_set* convert_ctable2hconv( canonical_table* ctable ) {
	huffman_conv_set* conv_set;
	int* llen;
	int* lval;
	int* vptr;
	
	
	// special case: only one symbol in table
	if ( ctable->n_symbols == 1 ) {
		conv_set = ( huffman_conv_set* ) calloc( 1, sizeof( huffman_conv_set ) );
		if ( conv_set == NULL ) return NULL;
		conv_set->h = ( huffman_conv** ) calloc( 1, sizeof( huffman_conv* ) );
		if ( conv_set->h == NULL ) { free( conv_set ); return NULL; }
		conv_set->h[0] = ( huffman_conv* ) calloc( 1, sizeof( huffman_conv ) );
		if ( conv_set->h[0] == NULL ) {	free( conv_set->h[0] );
			free( conv_set->h ); free( conv_set ); return NULL; }
		conv_set->h[0]->len = 0;
		conv_set->h[0]->val = (int) ctable->val_ptr[0]; 
		conv_set->h[0]->ext = NULL;
		conv_set->hwidth = 0;
		return conv_set;
	}
	
	// make a local copy of values and lengths array
	llen = (int*) calloc( ctable->max_clen + 1, sizeof( int ) );
	lval = (int*) calloc( ctable->n_symbols, sizeof( int ) );
	if ( ( llen == NULL ) || ( lval == NULL ) ) {
		if ( llen != NULL ) free( llen );
		if ( lval != NULL ) free( lval );
		return NULL; 
	}
	if ( ctable->len_ptr != NULL ) { // standard case
		for ( int i = 0; i < ctable->max_clen; i++ ) llen[ i + 1 ] = (int) ctable->len_ptr[ i ];
		for ( int i = 0; i < ctable->n_symbols; i++ ) lval[ i ] = (int) ctable->val_ptr[ i ];
	} else { // incompressible case
		llen[ 8 ] = 256;
		for ( int i = 0; i < 256; i++ ) lval[ i ] = i;
	}
	vptr = lval;
	
	// all the heavy lifing is done in a recursive function
	conv_set = convert_ctable2hconv_rec( llen, &vptr, ctable->max_clen );
	
	// free memory
	free( llen );
	free( lval );
	
	
	return conv_set;
}


/* -----------------------------------------------
	cleanup hcode table
	----------------------------------------------- */
void cleanup_hcode( huffman_code** table ) {
	for ( int i = 0; i < 256; i++ )
		if ( table[i] != NULL ) free( table[i] );
	free( table );
}


/* -----------------------------------------------
	cleanup hconv table
	----------------------------------------------- */
void cleanup_hconv( huffman_conv_set* conv_set ) {
	huffman_conv* tconv;
	huffman_conv** table = conv_set->h;
	int r = 1 << conv_set->hwidth;
	int c = 1;
	
	for ( tconv = table[0]; c < r; c++ ) {
		if ( tconv != table[c] ) {
			if ( tconv->ext != NULL ) cleanup_hconv( (huffman_conv_set*) tconv->ext );
			free( tconv );
			tconv = table[c];
		}			
	}
	free( tconv );
	free( table );
	free( conv_set );
}


/* -----------------------------------------------
	constructor for huffman reader class
	----------------------------------------------- */
	
huffman_reader::huffman_reader( unsigned char* data, int size )
{
	// WARNING: NO error checks in this class!
	// init bitreader
	bit_reader = new abitreader( data, size );
	// fill the bit buffer for the first time
	bit_buffer = bit_reader->read( BIT_BUFFER_SIZE );
}


/* -----------------------------------------------
	destructor for huffman reader class
	----------------------------------------------- */
	
huffman_reader::~huffman_reader( void )
{
	// close bitreader
	delete( bit_reader );
}


/* -----------------------------------------------
	decode one symbol from bitstream
	----------------------------------------------- */
	
int huffman_reader::decode_symbol( huffman_conv_set* conv_set )
{
	huffman_conv* conv;
	
	
	do {
		conv = conv_set->h[ bit_buffer >> ( BIT_BUFFER_SIZE - conv_set->hwidth ) ];
		conv_set = (huffman_conv_set*) conv->ext;
		advance_bitstream( conv->len );
	} while ( conv_set != NULL );
	
	
	return conv->val;
}


/* -----------------------------------------------
	bit reader function for n bit
	----------------------------------------------- */
	
unsigned int huffman_reader::read_bits( int n )
{
	unsigned int bits;
	
	// read bits
	bits = ( n <= BIT_BUFFER_SIZE ) ?
		bit_buffer >> (BIT_BUFFER_SIZE-n) :
		( bit_buffer << (n-BIT_BUFFER_SIZE) ) | bit_reader->read( (n-BIT_BUFFER_SIZE) );
	// refill the buffer
	advance_bitstream( n );
	
	return bits;
}


/* -----------------------------------------------
	bit reader function for one bit
	----------------------------------------------- */
	
unsigned char huffman_reader::read_bit( void )
{
	unsigned char bit;
	
	// read one bit
	bit = bit_buffer >> (BIT_BUFFER_SIZE-1);
	// refill the buffer
	advance_bitstream_1();
	
	return bit;
}


/* -----------------------------------------------
	return current byte position
	----------------------------------------------- */
	
int  huffman_reader::getpos( void )
{
	// ugly, but it works!
	return bit_reader->getpos() - ( BIT_BUFFER_SIZE + ( 8 - bit_reader->getbitp() ) ) / 8;
}


/* -----------------------------------------------
	refill bit buffer utility function
	----------------------------------------------- */
	
inline void huffman_reader::advance_bitstream( int n )
{
	// refill the buffer
	bit_buffer = ( n >= BIT_BUFFER_SIZE ) ?
		bit_reader->read( BIT_BUFFER_SIZE ) :
		( ( bit_buffer << n ) | bit_reader->read( n ) ) & ( ( 1 << BIT_BUFFER_SIZE ) - 1 );
}


/* -----------------------------------------------
	refill bit buffer utility function (1 bit)
	----------------------------------------------- */
	
inline void huffman_reader::advance_bitstream_1( void )
{
	// refill the buffer
	bit_buffer = ( ( bit_buffer << 1 ) | bit_reader->read_bit() ) &
		( ( 1 << BIT_BUFFER_SIZE ) - 1 );
}


/* -----------------------------------------------
	constructor for huffman writer class
	----------------------------------------------- */
	
huffman_writer::huffman_writer( int adds )
{
	// WARNING: NO error checks in this class!
	// init bitwriter - recommended value: 5MB
	if ( adds == 0 ) adds = 5 * 1024 * 1024;
	bit_writer = new abitwriter( adds );
}


/* -----------------------------------------------
	destructor for huffman writer class
	----------------------------------------------- */
	
huffman_writer::~huffman_writer( void )
{
	// close bitwriter
	delete( bit_writer );
}


/* -----------------------------------------------
	encode one symbol to bitstream
	----------------------------------------------- */
	
void huffman_writer::encode_symbol( huffman_code** hcodes, int symbol )
{
	huffman_code* hcode;	
	
	// find correct code, encode to bitstream
	hcode = hcodes[symbol];
	write_bits( hcode->code, hcode->len );
}


/* -----------------------------------------------
	bit writer function for n bit
	----------------------------------------------- */
	
void huffman_writer::write_bits( unsigned int bits, int n )
{
	// write bits
	bit_writer->write( bits, n );
}


/* -----------------------------------------------
	bit writer function for 1 bit
	----------------------------------------------- */
	
void huffman_writer::write_bit( unsigned char bit )
{
	// write bit
	bit_writer->write_bit( bit );
}


/* -----------------------------------------------
	return data pointer (when finished)
	----------------------------------------------- */
	
unsigned char* huffman_writer::getptr( void )
{
	return ( getpos() ) ? bit_writer->getptr() : NULL;
}


/* -----------------------------------------------
	return current byte position
	----------------------------------------------- */
	
int huffman_writer::getpos( void )
{
	return bit_writer->getpos();
}
