#include <stdlib.h>
#include <string.h>
#include "bitops.h"
#include "huffctable.h"

// maximum code length, must be 1 <= X < sizeof( int )
#define MAX_CLEN ( (int) ( (sizeof(int)*8) - 1 ) )


struct huffman_node {
	unsigned int count;
	// int clen;
	int val;
	huffman_node* next[2];
};

static void quicksort_nodes( huffman_node** lbound, huffman_node** rbound );
static void count_lengths( huffman_node* node, unsigned char* clen );
static void cleanup_tree( huffman_node* node );


/* -----------------------------------------------
	build canonical huffman table
	----------------------------------------------- */	
canonical_table* build_ctable( iostream* stream ) {
	unsigned int counts[ 256 ] = { 0 };
	unsigned char v;
	
	
	// count symbol occurences
	stream->rewind();
	while ( stream->read( &v, 1, 1 ) == 1 )
		counts[ (int) v ]++;
	
	
	return build_ctable( counts );
}


/* -----------------------------------------------
	build canonical huffman table
	----------------------------------------------- */	
canonical_table* build_ctable( unsigned int* counts ) {
	unsigned char clen[ 256 + 1 ] = { 0 };
	canonical_table* ctable;
	huffman_node** nodes;
	huffman_node** tree;
	huffman_node* n;
	int nn, ml;
	int i, j;
	
	
	// build nodes
	nodes = ( huffman_node** ) calloc( 256+1, sizeof( huffman_node* ) );
	if ( nodes == NULL ) return NULL;
	nodes[ 256 ] = NULL;
	for ( i = 0; i < 256; i++ ) {
		n = ( huffman_node* ) calloc( 1, sizeof( huffman_node ) );
		if ( n == NULL ) return NULL;
		n->count = 0;
		n->val = i;
		n->next[0] = NULL;
		n->next[1] = NULL;
		n->count = counts[ i ];
		nodes[i] = n;
	}
	
	// quicksort by counts
	quicksort_nodes( nodes, nodes + 255 );
	
	// remove zero count nodes (at least one node has to be left over)
	for ( i = 255; i > 0; i-- ) {
		if ( nodes[i]->count > 0 ) break;
		free( nodes[ i ] );
		nodes[i] = NULL;
	}	
	nn = i + 1; // number of remaining nodes (must be >= 1)
	
	// make a copy of the current nodes array
	tree = ( huffman_node** ) calloc( nn + 1, sizeof( huffman_node* ) );
	if ( tree == NULL ) return NULL;
	tree[ nn ] = NULL;
	memcpy( tree, nodes, nn * sizeof( huffman_node* ) );
	
	// build huffman tree
	for ( i = nn - 1; i > 0; i-- ) {
		// build new node combining two nodes with smallest count
		n = ( huffman_node* ) calloc( 1, sizeof( huffman_node ) );
		if ( n == NULL ) return NULL;
		n->count = tree[i]->count + tree[i-1]->count;
		n->val = -1;
		n->next[0] = tree[i-1];
		n->next[1] = tree[i];
		// replace old nodes, sort new node to its place
		for ( j = i - 1; ( j > 0 ) && ( n->count > tree[j-1]->count ) ; j-- )
			tree[j] = tree[j-1];
		tree[i] = NULL;
		tree[j] = n;
	}
	
	// count code lengths using tree
	count_lengths( (*tree), clen );
	if ( clen[0] == 1 ) {
		clen[0] = 0;
		clen[1] = 1;
	}
	
	// find max code length & limit code length if needed
	for ( ml = 256; ( clen[ml] == 0 ) && ( ml > 0 ); ml-- );
	while ( ml > MAX_CLEN ) {
		while ( clen[ml] > 0 ) {
			// find end node at least 2 below
			for ( i = ml - 2; clen[i] == 0; i-- );
			// replace a tree node with a value
			clen[ml]   -= 2;
			clen[ml-1] += 1;
			// replace a value with a tree node
			clen[i+1]  += 2;
			clen[i]    -= 1;
		}
		while( clen[--ml] == 0 );
	}
	if ( ml == 0 ) ml = ( nn == 256 ) ? 8 : 1;
	
	// build canonical table
	ctable = ( canonical_table* ) calloc ( 1, sizeof( canonical_table ) );
	if ( ctable == NULL ) return NULL;
	if ( ( nn == 256 ) && ( ml == 8 ) ) { // incompressible case
		// also handle this case in huffcoder routines!
		ctable->max_clen = 8;
		ctable->n_symbols = 256;
		ctable->total_size = 2;
		ctable->len_ptr = NULL;
		ctable->val_ptr = NULL;
		ctable->data = ( unsigned char* ) calloc( 2, 1 );
		ctable->counts = ( unsigned int* ) calloc( nn, sizeof( int ) );
		if ( ( ctable->data == NULL ) || ( ctable->counts == NULL ) ) return NULL;
		for ( i = 0; i < 256; i++ ) ctable->counts[ i ] = nodes[i]->count;
		ctable->data[0] = 0xFF;
		ctable->data[1] = 0xFF;
	} else { // generic routine
		ctable->max_clen = ml;
		ctable->n_symbols = nn;
		ctable->total_size = 1 + ml + nn;
		ctable->data = ( unsigned char* ) calloc( ctable->total_size, 1 );
		ctable->counts = ( unsigned int* ) calloc( nn, sizeof( int ) );
		if ( ( ctable->data == NULL ) || ( ctable->counts == NULL ) ) return NULL;
		ctable->len_ptr = ctable->data + 1;
		ctable->val_ptr = ctable->data + 1 + ml;
		ctable->data[0] = ml;
		memcpy( ctable->len_ptr, clen + 1, ml );
		for ( i = 0; i < nn; i++ ) {
			ctable->val_ptr[ i ] = nodes[i]->val;
			ctable->counts[ i ] = nodes[i]->count;
		}
	}
	
	
	// free memory
	cleanup_tree( (*tree) );
	free( tree );
	free( nodes );
	
	
	return ctable;
}


/* -----------------------------------------------
	read canonical huffman table from file
	----------------------------------------------- */
canonical_table* read_ctable( iostream* stream ) {
	canonical_table* ctable;
	unsigned char ml;
	int nn;
	int t;
	
	
	// allocate memory
	ctable = (canonical_table*) calloc( 1, sizeof( canonical_table ) );
	if ( ctable == NULL ) return NULL;
	ctable->counts = NULL;
	
	// read first byte, allocate memory
	stream->read( &ml, 1, 1 );
	ctable->data = (unsigned char*) calloc( 1 + ml, 1 );
	if ( ctable->data == NULL ) {
		free( ctable );
		return NULL;
	}
	
	// read file in 2 parts
	ctable->data[0] = ml;
	stream->read( ctable->data + 1, 1, 1 );
	if ( ( ctable->data[0] == 0xFF ) && ( ctable->data[1] == 0xFF ) ) { // special case (easy)
		ctable->max_clen = 8;
		ctable->n_symbols = 256;
		ctable->total_size = 2;
		ctable->len_ptr = NULL;
		ctable->val_ptr = NULL;
	} else { // standard case (read table in 2 parts)
		if ( stream->read( ctable->data + 2, 1, ml - 1 ) != ml - 1 ) {
			free( ctable->data );
			free( ctable );
			return NULL;
		}
		for ( nn = 0, t = 0; t < ml; nn += ctable->data[++t] );
		if ( nn > 256 ) {
			free( ctable->data );
			free( ctable );
			return NULL;
		}
		ctable->data = (unsigned char*) realloc( ctable->data, ml + nn + 1 );
		if ( ctable->data == NULL ) return NULL;
		if ( stream->read( ctable->data + 1 + ml, 1, nn ) != nn ) {
			free( ctable->data );
			free( ctable );
			return NULL;
		}
		ctable->max_clen = ml;
		ctable->n_symbols = nn;
		ctable->total_size = 1 + ml + nn;
		ctable->len_ptr = ctable->data + 1;
		ctable->val_ptr = ctable->data + 1 + ml;
	}
	
	
	return ctable;
}


/* -----------------------------------------------
	free memory from ctable
	----------------------------------------------- */	
void cleanup_ctable( canonical_table* ctable ) {
	if ( ctable != NULL ) {
		if ( ctable->counts != NULL ) free( ctable->counts );
		if ( ctable->data != NULL ) free( ctable->data );
		free( ctable );
	}
}


/* -----------------------------------------------
	quicksort huffman nodes by counts (helper)
	----------------------------------------------- */	
void quicksort_nodes( huffman_node** lbound, huffman_node** rbound ) {
	huffman_node** l;
	huffman_node** r;
	huffman_node* s;
	unsigned int p;
	
	
	// finished if this happens
	if ( lbound >= rbound ) return;
	
	// --- step 1: divide... ---
	l = lbound;
	r = rbound - 1;
	p = (*rbound)->count;
	
	// search for new pivot
	while ( true ) {
		// from left: search element smaller than pivot
		for ( ; ( (*l)->count >= p ) && ( l < rbound ); l++ );
		// from right: search element bigger than pivot
		for ( ; ( (*r)->count <= p ) && ( r > lbound ); r-- );
		// swap if l still on the left, r on the right
		if ( l < r ) {
			s = (*l); (*l) = (*r); (*r) = s;
		} else break; // ...break otherwise
	}
	
	// swap pivot data
	if ( (*l)->count < p ) {
		s = (*l); (*l) = (*rbound); (*rbound) = s;
	}
	
	// --- step 2: ... and conquer! ---
	quicksort_nodes( lbound, l - 1 );
	quicksort_nodes( l + 1, rbound );
	
	
	return;
}


/* -----------------------------------------------
	count code lengths of end nodes (helper)
	----------------------------------------------- */
static void count_lengths( huffman_node* node, unsigned char* clen ) {
	if ( ( node->next[0] == NULL ) && ( node->next[1] == NULL ) ) (*clen)++;
	else {
		clen++;
		if ( node->next[0] != NULL ) count_lengths( node->next[0], clen );
		if ( node->next[1] != NULL ) count_lengths( node->next[1], clen );
	}
}


/* -----------------------------------------------
	free all nodes in tree (helper)
	----------------------------------------------- */	
static void cleanup_tree( huffman_node* node ) {
	if ( node->next[0] != NULL ) cleanup_tree( node->next[0] );
	if ( node->next[1] != NULL ) cleanup_tree( node->next[1] );
	free( node );
}
